/* Copyright (c) 2019-2021, Sonos, Inc.
 *
 * SPDX-License-Identifier: GPL-2.0
 *
 */
#ifndef _SIMPLE80211_H_
#define _SIMPLE80211_H_
#include <linux/ieee80211.h>
#ifdef SONOS_ARCH_ATTR_SUPPORTS_WOSSID
#include <linux/time.h>
#endif
#include "simple80211_ioctl.h"

/* FIXME: The bridge name should be decided by upper layer. */
#define SIMPLE80211_BRIDGE_NAME "br0"

#define SIMPLE80211_OP_MODE_NORMAL 0

struct simple80211_info;

#define SIMPLE80211_STA_STATE_IDLE                 0
#define SIMPLE80211_STA_STATE_SCAN                 1
#define SIMPLE80211_STA_STATE_SCAN_DONE            2
#define SIMPLE80211_STA_STATE_WAIT_FOR_AP          3
#define SIMPLE80211_STA_STATE_WAIT_FOR_AP_DONE     4
#define SIMPLE80211_STA_STATE_AUTH                 5
#define SIMPLE80211_STA_STATE_AUTH_SHARED          6
#define SIMPLE80211_STA_STATE_AUTH_CHALLENGE       7
#define SIMPLE80211_STA_STATE_AUTH_DONE            8
#define SIMPLE80211_STA_STATE_ASSOC                9
#define SIMPLE80211_STA_STATE_ASSOC_DONE           10
#define SIMPLE80211_STA_STATE_4WAY_HANDSHAKE       11
#define SIMPLE80211_STA_STATE_WAIT_KEY             12
#define SIMPLE80211_STA_STATE_CONNECTED            13

/* Enum values should be consistent with station_ioctl.h */
enum simple80211_sta_mode {
    SIMPLE80211_STA_DISABLED,
    SIMPLE80211_STA_ENABLED,
    SIMPLE80211_STA_CREDCHECK,
    SIMPLE80211_STA_CREDGOOD,
    SIMPLE80211_STA_CREDBAD,
};

enum simple80211_cmd_types {
    SIMPLE80211_INIT_INFO,
    SIMPLE80211_BR_UPLINK,
    SIMPLE80211_RF_READY,
    SIMPLE80211_SCAN_EVENT,
    SIMPLE80211_IOCTL,
    SIMPLE80211_ADD_KEY,
    SIMPLE80211_TXRX_FILTER,
    SIMPLE80211_STA_GET_STATE,
    SIMPLE80211_SUSPEND,
    SIMPLE80211_RESUME,
    SIMPLE80211_CMD_END,
};

/* Enum values should be consistent with station_ioctl.h */
enum simple80211_sta_event {
    SIMPLE80211_STA_CONNECTED       = 700,
    SIMPLE80211_STA_DISCONNECTED    = 701,
    SIMPLE80211_STA_EVENT_CREDGOOD  = 705,
    SIMPLE80211_STA_EVENT_CREDBAD   = 706,
    SIMPLE80211_STA_EVENT_SUPP_STOP = 906,
};

enum simple80211_sta_event_connected_data {
    SIMPLE80211_STA_EVENT_DATA_CLEAR = 1,
    SIMPLE80211_STA_EVENT_DATA_WEP   = 2,
    SIMPLE80211_STA_EVENT_DATA_AES   = 3,
    SIMPLE80211_STA_EVENT_DATA_TKIP  = 4
};

enum simple80211_scan_event {
    SIMPLE80211_SCAN_EVENT_PRESCAN,
    SIMPLE80211_SCAN_EVENT_SCANNING,
    SIMPLE80211_SCAN_EVENT_SCAN_DONE,
    SIMPLE80211_SCAN_EVENT_SCAN_ABORTED,
};

/* Must make sure these match what the user-space expects. */
enum simple80211_resume_reason {
    SIMPLE80211_RESUME_NONE = 0,
    SIMPLE80211_RESUME_NORMAL,
    SIMPLE80211_RESUME_BCN_LOSS,
    SIMPLE80211_RESUME_KEEP_ALIVE,
    SIMPLE80211_RESUME_DECRYPT,
    SIMPLE80211_RESUME_DEAUTH,
    SIMPLE80211_RESUME_DISASSOC,
    SIMPLE80211_RESUME_WOSSID,
};

/* Statistics events */
enum stat_event_ID_t {
	STAT_EVENT_BAD_CRED,
	STAT_EVENT_DISCONNECT,
	STAT_EVENT_CONNECT
};

struct simple80211_cmd_uplink {
    struct net_device *ndev;
    unsigned char *mac;
    unsigned char add;
    unsigned int cipher_pairwise;
    unsigned int cipher_group;
    char ssid[IEEE80211_MAX_SSID_LEN+1];
    int ssid_len;
    unsigned char is_psk;
};

struct simple80211_cmd_init_info {
    struct net_device *base_dev;
    void *lmac_ops;
    unsigned char is_init;
};

struct simple80211_cmd_scan_event {
    unsigned int event_type;
    int skip_action;
};

struct simple80211_cmd_ioctl {
    unsigned long *args;
};

#ifdef WLAN_CIPHER_SUITE_NONE
#error "WLAN_CIPHER_SUITE_NONE is not expected to be defined."
#else
#define WLAN_CIPHER_SUITE_NONE 0
#endif

#ifdef WLAN_CIPHER_SUITE_INVALID
#error "WLAN_CIPHER_SUITE_INVALID is not expected to be defined."
#else
#define WLAN_CIPHER_SUITE_INVALID -1
#endif

#ifndef IEEE80211_ADDR_LEN
#define IEEE80211_ADDR_LEN 6
#endif

/* Copied from atheros/driver/station_ioctl.h */
#define STA_STATE_FLAG_ENABLED      (1 << 0)
#define STA_STATE_FLAG_CONNECTED    (1 << 1)

#define STA_BEACON_IES_LEN 1000

#define WOSSID_MAX_SSID 16
#define WOSSID_WAKE_LOOP_THRESHOLD 3
#define WOSSID_WAKE_LOOP_SCAN_INTERVAL ((60 + 3) * 3) /* seconds */
#define WOSSID_WAKE_LOOP_SUSPEND_INTERVAL (10 * 60) /* seconds */

/* Derived from atheros/driver/station_ioctl.h */
struct simple80211_sta_state_info {
    unsigned long flags; /* High level view of some of the states */
    int  state;          /* No idea why we send this.  It is not wise to use it */
    int  freq;           /* Current freq, which may not be what you expect */
    int  ap_freq;        /* Freq AP is on.  0 if not connected */

    /* SSID/BSSID we last considered.  This may be non-empty even if disabled */
    int  ssid_len;
    char ssid[IEEE80211_MAX_SSID_LEN+1];
    char bssid[IEEE80211_ADDR_LEN];

    /* Information about the beacons the AP we last considered is sending */
    unsigned short beacon_int;
    unsigned short beacon_capa;
    unsigned int   beacon_ies_len;
    char           beacon_ies[STA_BEACON_IES_LEN];
};

struct simple80211_cmd_add_key {
    unsigned int cipher;
    unsigned char is_pair;
};

struct simple80211_cmd_txrx_filter {
    struct sk_buff *skb;
    unsigned char is_tx;
};

struct simple80211_cmd_power_state {
    uint8_t resume_reason;
};

struct simple80211_cmd {
    unsigned int cmd_type;
    union {
        struct simple80211_cmd_init_info init;
        struct simple80211_cmd_uplink uplink;
        struct simple80211_cmd_scan_event scan_event;
        struct simple80211_cmd_ioctl ioctl;
        struct simple80211_cmd_add_key key;
        struct simple80211_cmd_txrx_filter txrx_filter;
        struct simple80211_cmd_power_state power_state;
        struct simple80211_sta_state_info sta_info;
    } data;
};

struct simple80211_ssid {
    u_int8_t ssid[IEEE80211_MAX_SSID_LEN];
    u_int8_t ssid_len;
};

struct simple80211_channel {
    u_int16_t freq;
};

/* Used by the UMAC to describe the scan should be done by LMAC. */
struct simple80211_scan_request {
    u_int8_t start_scan;
    struct simple80211_channel *channels;
    u_int8_t channel_num;
    struct simple80211_ssid *ssids;
    u_int8_t ssid_num;
};

#define SCAN_FLAG_RUNNING          (1 << 0)
#define SCAN_FLAG_INIT_SCAN        (1 << 1)
#define SCAN_FLAG_CACHED           (1 << 2)
#define SCAN_FLAG_CALLER_WAITING   (1 << 3)
#define SCAN_FLAG_UMAC_SCAN        (1 << 4)
/* Scan related defines and enums */
struct simple80211_scan_info {
    u_int16_t scan_flags;
    struct simple80211_scan_results scan_results;
};

struct simple80211_shim {
    const struct simple80211_shim_ops *shim_ops;
    void *priv;
};

/* Types and structs used by functions called by UMAC into LMAC. */
struct simple80211_shim_ops {
    void (*credcheck_reset)(struct simple80211_info *info);
    int (*get_beacon_capabilities)(struct simple80211_info *info, unsigned int *beacon_capabilities);
    int (*get_beacon_interval)(struct simple80211_info *info, unsigned int *beacon_interval);
    int (*get_channel)(struct simple80211_info *info, unsigned int *channel);
    int (*get_link_stats)(struct simple80211_info *info,
                          struct simple80211_link_stats_info *pInfoLocal,
                          unsigned long thresholdTxPktsMin,
                          unsigned long thresholdLatencyMs);
    int (*link_quality_polling)(struct simple80211_info *info, unsigned int enable);
    void (*pwr_resume)(struct simple80211_info *info);
    void (*pwr_suspend)(struct simple80211_info *info);
    int (*scan)(struct simple80211_info *info, struct simple80211_scan_request *shim_req);
    int (*scan_abort)(struct simple80211_info *info);
    int (*scan_clear)(struct simple80211_info *info);
    int (*scan_save_results)(struct simple80211_info *info);
    int (*scan_start)(struct simple80211_info *info, unsigned long *args);
    int (*send_stat_event)(struct simple80211_info *info, enum stat_event_ID_t stat_event_ID,
                           unsigned int count);
    int (*set_channel)(struct simple80211_info *info, unsigned int channel);
    void (*start_wossid)(struct simple80211_info *info);
    void (*stop_wossid)(struct simple80211_info *info);
};

struct simple80211_txrx_filter_node {
    struct list_head list;
    struct simple80211_txrx_filter filter;
};

struct simple80211_info {
    struct net_device *base_dev;
    struct simple80211_shim shim;
    unsigned char     uplink_enabled;
    unsigned char     valid;
    unsigned char     open_mode;
    unsigned char     sta_mode;
    unsigned char     sta_state;
    unsigned char     sta_scan_count;
    unsigned int      sta_cipher_pairwise;
    unsigned int      sta_cipher_group;
    unsigned int      sta_connection_failures;
    struct list_head  txrx_filter_list;
    char              ssid[IEEE80211_MAX_SSID_LEN+1];
    int               ssid_len;
    unsigned char     bssid[IEEE80211_ADDR_LEN];
    spinlock_t        filter_lock;
    spinlock_t        info_lock;

    /* This is used to indicate current operation mode, such as normal station,
    or as a station of HT primary. */
    uint8_t operation_mode;
    struct simple80211_scan_info scan_info;
#ifdef SONOS_ARCH_ATTR_SUPPORTS_SUSPEND
    u_int32_t ip_addr;
    uint8_t   resume_reason;
#ifdef SONOS_ARCH_ATTR_SUPPORTS_WOSSID
    struct simple80211_ssid wossid_list[WOSSID_MAX_SSID];
    struct timeval wossid_suspend_time;
    struct timeval wossid_resume_time;
    int wossid_wake_loop_count;
    uint8_t wossid_started;
    uint8_t wossid_list_len;
#endif
#endif
};

/* Externally visible functions */
extern int simple80211_handle_cmd(struct simple80211_info *info,
                                  struct simple80211_cmd *cmd);
#endif /* _SIMPLE80211_H_ */
