/* Copyright (c) 2019-2021, Sonos, Inc.
 *
 * SPDX-License-Identifier: GPL-2.0
 *
 */
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/if_bridge.h>
#include <linux/skbuff.h>
#include <asm/uaccess.h>
#include <linux/rtnetlink.h>
#include <net/rtnetlink.h>
#include <net/netlink.h>
#include <net/sock.h>
#include <net/cfg80211.h>
#include <net/ip.h>
#include <net/ipv6.h>
#include "simple80211.h"

typedef int (*nl_put_data)(struct simple80211_info *info, struct sk_buff *skb,
                           int event, void *data, int data_len);
#define DST_GROUP dst_group
#ifdef SONOS_ARCH_ATTR_SUPPORTS_MT7663
#include "mt7663_shim.h"
#endif

#define NLMSG_PUT(skb, portid, seq, type, len)                 \
({                                                             \
    if (unlikely(skb_tailroom(skb) < nlmsg_total_size(len))) { \
        goto nlmsg_failure;                                    \
    }                                                          \
    __nlmsg_put(skb, portid, seq, type, len, 0);               \
})

#define RTA_PUT(_skb, _attrtype, _attrlen, _data)              \
  if(nla_put((_skb), (_attrtype), (_attrlen), (_data)) < 0) {  \
      goto rtattr_failure;                                     \
  }

#define nl_alloc(_size) nlmsg_new(_size, GFP_ATOMIC)
#define nl_broadcast(_skb) \
    NETLINK_CB(skb).DST_GROUP = RTMGRP_Rincon ; \
    rtnl_notify(_skb, sock_net(rtnl), 0, RTMGRP_Rincon, NULL, GFP_ATOMIC);

void simple80211_free_txrx_filters(struct simple80211_info *info);

static int simple80211_in_credcheck(struct simple80211_info *info)
{
    return (info->sta_mode >= SIMPLE80211_STA_CREDCHECK);
}

int simple80211_sta_event(struct simple80211_info *info, struct sk_buff *skb,
                          int event, void *data, int data_len)
{
    struct nlmsghdr *nlh;
    unsigned char *b = skb_tail_pointer(skb);

    nlh = NLMSG_PUT(skb, 0, 0, RWM_SUPPLICANT, 0);
    RTA_PUT(skb, RWA_DEV_NAME, IFNAMSIZ, info->base_dev->name);
    RTA_PUT(skb, RWA_SUPPLICANT_EVENT, sizeof(event), &event);
    if (data && data_len) {
        RTA_PUT(skb, RWA_SUPPLICANT_EVENT_DATA, data_len, data);
    }

    nlh->nlmsg_len = skb_tail_pointer(skb) - b;
    return skb->len;
nlmsg_failure:
rtattr_failure:
    skb_trim(skb, b - skb->data);
    return -EINVAL;
}

int simple80211_renew_dhcp_event(struct simple80211_info *info, struct sk_buff *skb,
                                 int event, void *data, int data_len)
{
    struct nlmsghdr *nlh;
    unsigned char *b = skb_tail_pointer(skb);

    nlh = NLMSG_PUT(skb, 0, 0, RWM_RENEW_DHCP, 0);
    nlh->nlmsg_len = skb_tail_pointer(skb) - b;
    (void)info;
    (void)event;
    (void)data;
    (void)data_len;

    return skb->len;
nlmsg_failure:
    skb_trim(skb, b - skb->data);
    return -EINVAL;
}

static int simple80211_nl_event(struct simple80211_info *info, int event,
                                void *data, int len, nl_put_data data_func)
{
    struct sk_buff *skb;
    int size = NLMSG_SPACE(1024);
    struct sock *rtnl = NULL;
    int ret;

    skb = nl_alloc(size);
    if (!skb) {
        netlink_set_err(rtnl, 0, RTMGRP_Rincon, ENOBUFS);
        printk("%s: failed netlink alloc\n", __func__);
        return -ENOMEM;
    }
    ret = data_func(info, skb, event, data, len);
    if (ret < 0) {
        kfree_skb(skb);
        netlink_set_err(rtnl, 0, RTMGRP_Rincon, EINVAL);
        printk("%s: failed netlink event\n", __func__);
        return ret;
    }
    nl_broadcast(skb);
    return 0;
}

int simple80211_handle_connected(struct simple80211_info *info)
{
    int ret = 0;
    int conn_data = 0;
    int send_cred_good = 1;
    info->sta_state = SIMPLE80211_STA_STATE_CONNECTED;
    switch (info->sta_cipher_pairwise) {
        case WLAN_CIPHER_SUITE_WEP40:
        case WLAN_CIPHER_SUITE_WEP104:
            conn_data = SIMPLE80211_STA_EVENT_DATA_WEP;
            if (!simple80211_in_credcheck(info)) {
                /* WEP - Only send good cred event if it's credcheck. */
                send_cred_good = 0;
            }
            break;
        case WLAN_CIPHER_SUITE_NONE:
            conn_data = SIMPLE80211_STA_EVENT_DATA_CLEAR;
            break;
        case WLAN_CIPHER_SUITE_TKIP:
            conn_data = SIMPLE80211_STA_EVENT_DATA_TKIP;
            break;
        case WLAN_CIPHER_SUITE_CCMP:
            conn_data = SIMPLE80211_STA_EVENT_DATA_AES;
            break;
        default:
            ret = -EINVAL;
            printk("ASSERT: %s - Unknown cipher %u\n", __func__, info->sta_cipher_pairwise);
    }

    if (send_cred_good) {
        simple80211_nl_event(info, SIMPLE80211_STA_EVENT_CREDGOOD,
                             NULL, 0, simple80211_sta_event);
    }

    if (!simple80211_in_credcheck(info)) {
        simple80211_nl_event(info, SIMPLE80211_STA_CONNECTED,
                             &conn_data, sizeof(conn_data), simple80211_sta_event);
        printk("CONNECTED - waiting for IP\n");
    } else {
        printk("CONNECTED\n");
    }

    info->shim.shim_ops->send_stat_event(info, STAT_EVENT_CONNECT, 0);
    return ret;
}

void simple80211_reset_connection_info(struct simple80211_info *info)
{
    info->sta_state = SIMPLE80211_STA_STATE_IDLE;
    info->sta_cipher_pairwise = WLAN_CIPHER_SUITE_INVALID;
    info->sta_cipher_group = WLAN_CIPHER_SUITE_INVALID;
    info->ssid_len = 0;
    info->shim.shim_ops->link_quality_polling(info, 0);
}

void simple80211_credcheck_reset(struct simple80211_info *info)
{
    if (info->shim.shim_ops->credcheck_reset) {
        info->shim.shim_ops->credcheck_reset(info);
    }
}

int simple80211_uplink(struct simple80211_info *info,
                       struct simple80211_cmd *cmd)
{
    struct ifreq rq;
    unsigned long args[4];
    struct net_device *br_dev = NULL;
    mm_segment_t oldfs;
    int rc = 0;

    spin_lock_bh(&info->info_lock);
    if (simple80211_in_credcheck(info)) {
        if (!cmd->data.uplink.add) {
            int send_bad_cred_event = 1;

            if (cmd->data.uplink.is_psk) {
                /* Try up to three BSSIDs before reporting failure (SWPBL-158946) */
                if (++info->sta_connection_failures < 3) {
                    printk("%s: sta_connection_failures: %u\n", __func__, info->sta_connection_failures);
                    send_bad_cred_event = 0;
                } else {
                    info->sta_connection_failures = 0;
                }
            } else {
                if (info->sta_scan_count < 3) {
                    printk("%s: sta_scan_count: %d\n", __func__, info->sta_scan_count);
                    send_bad_cred_event = 0;
                }
            }

            if (send_bad_cred_event) {
                printk("%s: Sending bad cred event\n", __func__);
                info->sta_scan_count = 0;
                simple80211_nl_event(info, SIMPLE80211_STA_EVENT_CREDBAD,
                                     NULL, 0, simple80211_sta_event);
                simple80211_credcheck_reset(info);
            }
            simple80211_reset_connection_info(info);
            spin_unlock_bh(&info->info_lock);
            return 0;
        }
    }

    if (info->sta_state == SIMPLE80211_STA_STATE_CONNECTED && !cmd->data.uplink.add) {
        info->shim.shim_ops->send_stat_event(info, STAT_EVENT_DISCONNECT, 0);
    }

    if (info->sta_state == SIMPLE80211_STA_STATE_ASSOC_DONE &&
        !cmd->data.uplink.add &&
        cmd->data.uplink.cipher_pairwise != WLAN_CIPHER_SUITE_NONE) {
        info->shim.shim_ops->send_stat_event(info, STAT_EVENT_BAD_CRED, 0);
    }

    if (!cmd->data.uplink.add == !info->uplink_enabled) {
        if (!cmd->data.uplink.add && info->sta_mode == SIMPLE80211_STA_DISABLED) {
            /* If we get here, it implies that we have stale sta_state info left from credcheck. Just clear
             * them and return.
             */
            simple80211_reset_connection_info(info);
            rc = 0;
        } else {
            printk("%s: repeating operation add %d\n", __func__, !!cmd->data.uplink.add);
            rc = -1;
        }
        spin_unlock_bh(&info->info_lock);
        return rc;
    }

    if (cmd->data.uplink.add) {
        info->sta_state = SIMPLE80211_STA_STATE_ASSOC_DONE;
        info->shim.shim_ops->link_quality_polling(info, 1);
        if (cmd->data.uplink.ssid_len) {
            info->ssid_len = cmd->data.uplink.ssid_len;
            memcpy(info->ssid, cmd->data.uplink.ssid, cmd->data.uplink.ssid_len);
            memcpy(info->bssid, cmd->data.uplink.mac, IEEE80211_ADDR_LEN);
        }
    } else {
        simple80211_reset_connection_info(info);
        simple80211_nl_event(info, SIMPLE80211_STA_DISCONNECTED, NULL, 0, simple80211_sta_event);
    }
    if (info->sta_state == SIMPLE80211_STA_STATE_ASSOC_DONE &&
        cmd->data.uplink.cipher_pairwise == WLAN_CIPHER_SUITE_NONE) {
        info->sta_cipher_pairwise = cmd->data.uplink.cipher_pairwise;
        /* Connected and no cipher is needed. Send good cred event. */
        simple80211_handle_connected(info);
    }
    if (simple80211_in_credcheck(info)) {
        /* Do nothing to bridge. */
        spin_unlock_bh(&info->info_lock);
        return 0;
    }

    args[0] = (unsigned long)(cmd->data.uplink.add ? BRCTL_ADD_UPLINK : BRCTL_DEL_P2P_TUNNEL);
    args[1] = (unsigned long)cmd->data.uplink.ndev->ifindex;
    args[2] = (unsigned long)cmd->data.uplink.mac;
    args[3] = 0;

    br_dev = dev_get_by_name(&init_net, SIMPLE80211_BRIDGE_NAME);
    if (!br_dev) {
        printk("%s: can't find %s\n", __func__, SIMPLE80211_BRIDGE_NAME);
        spin_unlock_bh(&info->info_lock);
        return -1;
    }
    rq.ifr_data = (char*)args;
    strncpy(rq.ifr_name, br_dev->name, IFNAMSIZ - 1);
    rq.ifr_name[IFNAMSIZ - 1] = '\0';
    oldfs = get_fs();
    set_fs(KERNEL_DS);
    br_dev->netdev_ops->ndo_do_ioctl(br_dev, &rq, SIOCDEVPRIVATE);
    set_fs(oldfs);
    info->uplink_enabled = !!cmd->data.uplink.add;
    dev_put(br_dev);
    spin_unlock_bh(&info->info_lock);
    return 0;
}

/* This should be called by the code that allocates net_device for the actual wireless
 * base interface to initialize the simple80211_info.
 */
int simple80211_init_info(struct simple80211_info *info,
                          struct simple80211_cmd *cmd)
{
    if (cmd->data.init.is_init) {
        /* Initialize the simple80211_info */
        memset(info, 0, sizeof(struct simple80211_info));
        spin_lock_init(&info->info_lock);
        spin_lock_init(&info->filter_lock);

        spin_lock_bh(&info->info_lock);
        info->base_dev = cmd->data.init.base_dev;
        /* shim_init() should be linked to the right one in compile time. */
        shim_init(info, cmd->data.init.lmac_ops, 1);
        info->valid = 1;
        memset(&info->scan_info, 0, sizeof(info->scan_info));
        info->scan_info.scan_flags = 0;
        info->base_dev->flags |= IFF_ALLMULTI;
        simple80211_reset_connection_info(info);
        spin_lock_bh(&info->filter_lock);
        INIT_LIST_HEAD(&info->txrx_filter_list);
        spin_unlock_bh(&info->filter_lock);
        info->operation_mode = SIMPLE80211_OP_MODE_NORMAL;
        spin_unlock_bh(&info->info_lock);
    } else {
        /* Reset the simple80211_info */
        spin_lock_bh(&info->info_lock);
        simple80211_free_txrx_filters(info);
        info->valid = 0;
        shim_init(info, NULL, 0);
        spin_unlock_bh(&info->info_lock);
    }
    return 0;
}

int simple80211_ioctl_set_hhid(struct simple80211_info *info,
                               unsigned long *args)
{
    unsigned char hhid[IEEE80211_MAX_HHID_LEN + 1];
    unsigned long hhid_len = args[1];
    struct sk_buff *skb;
    int size;
    struct sock *rtnl = NULL;
    unsigned char *skb_tail;
    struct nlmsghdr *nlh;

    if (hhid_len > IEEE80211_MAX_HHID_LEN) {
        return -EINVAL;
    }
    memset(hhid, 0, sizeof(hhid));
    if (copy_from_user(hhid, (void *)args[2], hhid_len)) {
        return -EFAULT;
    }

    size = NLMSG_SPACE(1024);
    skb = nl_alloc(size);
    if (!skb) {
        netlink_set_err(rtnl, 0, RTMGRP_Rincon, ENOBUFS);
        return -ENOMEM;
    }

    skb_tail = skb_tail_pointer(skb);

    nlh = NLMSG_PUT(skb, 0, 0, RWM_HHID, 0);
    RTA_PUT(skb, RWA_DEV_NAME, IFNAMSIZ, info->base_dev->name);
    RTA_PUT(skb, RWA_HHID, hhid_len, hhid);
    nlh->nlmsg_len = skb_tail_pointer(skb) - skb_tail;

    nl_broadcast(skb);

    printk("%s: Set HHID to \"%s\"\n", info->base_dev->name, hhid);
    return 0;
nlmsg_failure:
rtattr_failure:
    skb_trim(skb, skb_tail - skb->data);
    kfree_skb(skb);
    netlink_set_err(rtnl, 0, RTMGRP_Rincon, EINVAL);
    return -EINVAL;
}

int simple80211_ioctl_set_channel(struct simple80211_info *info,
                                  unsigned long *args)
{
    unsigned int channel;
    channel = (args[1] >= 2412) ? ieee80211_frequency_to_channel(args[1]) : args[1];
    if (channel == 0) {
        return -EINVAL;
    }
    /* FIXME: Need to check whether channel is valid once get there. */
    return info->shim.shim_ops->set_channel(info, channel);
}

int simple80211_ioctl_has_wireless_cnx(struct simple80211_info *info,
                                       unsigned long *args)
{
    unsigned char uplink_enabled = 0;

    if (info->sta_mode != SIMPLE80211_STA_DISABLED) {
        uplink_enabled = info->uplink_enabled;
    }

    if (copy_to_user((void *)args[1], &uplink_enabled, sizeof(uplink_enabled))) {
        return -EFAULT;
    }

    return 0;
}

int simple80211_ioctl_get_link_stats(struct simple80211_info *info,
                                     unsigned long *args)
{
    unsigned int rc;
    struct simple80211_link_stats_info *p80211_link_stats_info;
    struct simple80211_link_stats_info *pLink_stats_info_local;

    p80211_link_stats_info = (struct simple80211_link_stats_info *) args[3];
    pLink_stats_info_local = kmalloc(sizeof(struct simple80211_link_stats_info), GFP_KERNEL);
    if (!pLink_stats_info_local) {
        printk(KERN_WARNING "Out of memory at %s:%d\n", __FILE__, __LINE__);
        return -ENOMEM;
    }
    memset(pLink_stats_info_local, 0, sizeof(struct simple80211_link_stats_info));

    info->shim.shim_ops->get_channel(info, &pLink_stats_info_local->freq);
    info->shim.shim_ops->get_link_stats(info, pLink_stats_info_local, args[1], args[2]);

    memcpy(pLink_stats_info_local->apStats.bssid, info->bssid,
           sizeof(pLink_stats_info_local->apStats.bssid));

    /* apStats->ssid can't hold \0. Use info->ssid_len for length. */
    pLink_stats_info_local->apStats.ssidLen = info->ssid_len < IEEE80211_SSID_LEN ? info->ssid_len : IEEE80211_SSID_LEN;
    memcpy(pLink_stats_info_local->apStats.ssid, info->ssid, info->ssid_len);

    rc = copy_to_user((void *) p80211_link_stats_info, pLink_stats_info_local,
              sizeof(struct simple80211_link_stats_info));
    rc = rc ? -EFAULT : 0;
    kfree(pLink_stats_info_local);
    return rc;
}

int simple80211_ioctl_get_channel(struct simple80211_info *info,
                                  unsigned long *args)
{
    unsigned int freq;
    int ret = info->shim.shim_ops->get_channel(info, &freq);
    if (ret) {
        return ret;
    }
    if (copy_to_user((void *)args[1], &freq, sizeof(freq))) {
        return -EFAULT;
    }
    return 0;
}

int simple80211_ioctl_sta_set_enable(struct simple80211_info *info,
                                     unsigned long *args)
{
    info->sta_mode = (unsigned char)args[1];
    switch (info->sta_mode) {
        case SIMPLE80211_STA_DISABLED: {
            simple80211_nl_event(info, SIMPLE80211_STA_EVENT_SUPP_STOP, NULL, 0, simple80211_sta_event);
            break;
        }
        case SIMPLE80211_STA_CREDGOOD: {
            simple80211_handle_connected(info);
            break;
        }
        case SIMPLE80211_STA_CREDBAD: {
            simple80211_nl_event(info, SIMPLE80211_STA_EVENT_CREDBAD, NULL, 0, simple80211_sta_event);
            break;
        }
        case SIMPLE80211_STA_CREDCHECK: {
            info->sta_scan_count = 0;
            break;
        }
        case SIMPLE80211_STA_ENABLED:
        default:
            break; /* do nothing */
    }
    return 0;
}

int simple80211_ioctl_add_txrx_filter(struct simple80211_info *info,
                                      unsigned long *args)
{
    struct simple80211_txrx_filter_node *filter_node;
    filter_node = (struct simple80211_txrx_filter_node *)kmalloc(sizeof(struct simple80211_txrx_filter_node), GFP_KERNEL);

    if (!filter_node) {
        return -ENOMEM;
    }
    if (copy_from_user(&filter_node->filter, (void *)args[1], sizeof(filter_node->filter))) {
        kfree(filter_node);
        return -EFAULT;
    }

    spin_lock_bh(&info->filter_lock);
    list_add_tail(&filter_node->list, &info->txrx_filter_list);
    spin_unlock_bh(&info->filter_lock);
    return 0;
}

void simple80211_free_txrx_filters(struct simple80211_info *info)
{
    struct simple80211_txrx_filter_node *filter_node, *tmp;
    spin_lock_bh(&info->filter_lock);
    list_for_each_entry_safe(filter_node, tmp, &info->txrx_filter_list, list) {
        list_del(&filter_node->list);
        kfree(filter_node);
    }
    spin_unlock_bh(&info->filter_lock);
}

/* This code is almost identical to simple80211_ioctl_sta_get_state().
 * The code will be merged in the next iteration. */
int simple80211_sta_get_state(struct simple80211_info *info,
                              struct simple80211_cmd *cmd)
{
    int iRC;
    unsigned int channel;
    unsigned int beacon_interval;
    unsigned int beacon_capabilities;
    struct simple80211_sta_state_info *pSta_info;

    pSta_info = &cmd->data.sta_info;
    memset(pSta_info, 0, sizeof(struct simple80211_sta_state_info));

    if (info->sta_mode != SIMPLE80211_STA_DISABLED) {
        pSta_info->flags |= STA_STATE_FLAG_ENABLED;

        if (SIMPLE80211_STA_STATE_CONNECTED == info->sta_state) {
            pSta_info->flags |= STA_STATE_FLAG_CONNECTED;
        }
    }

    pSta_info->state = info->sta_state;

    if (info->ssid_len) {
        pSta_info->ssid_len = info->ssid_len;
        memcpy(pSta_info->ssid, info->ssid, info->ssid_len);
        memcpy(pSta_info->bssid, info->bssid, IEEE80211_ADDR_LEN);
    }

    iRC = info->shim.shim_ops->get_beacon_interval(info, &beacon_interval);
    if (iRC) {
        beacon_interval = 0;
    }
    pSta_info->beacon_int = beacon_interval;

    iRC = info->shim.shim_ops->get_beacon_capabilities(info, &beacon_capabilities);
    if (iRC) {
        beacon_capabilities = 0;
    }
    pSta_info->beacon_capa = beacon_capabilities;

    iRC = info->shim.shim_ops->get_channel(info, &channel);
    if (iRC) {
        channel = 0;
    }
    pSta_info->freq = pSta_info->ap_freq = channel;

    return 0;
}

/* For now only a subset of the simple80211_sta_state_info struct are used. Below are the ones get used:
 * - flags
 * - ssid_len
 * - ssid
 * - bssid
 */
int simple80211_ioctl_sta_get_state(struct simple80211_info *info,
                                    unsigned long *args)
{
    unsigned int channel;
    unsigned int beacon_interval;
    unsigned int beacon_capabilities;
    int iRC;
    struct simple80211_sta_state_info *pSta_info;

    pSta_info = kmalloc(sizeof(struct simple80211_sta_state_info), GFP_KERNEL);

    if (!pSta_info) {
        printk(KERN_WARNING "Out of memory at %s:%d\n", __FILE__, __LINE__);
        return -ENOMEM;
    }

    memset(pSta_info, 0, sizeof(struct simple80211_sta_state_info));

    if(info->sta_mode != SIMPLE80211_STA_DISABLED) {
        pSta_info->flags |= STA_STATE_FLAG_ENABLED;

        if (SIMPLE80211_STA_STATE_CONNECTED == info->sta_state) {
            pSta_info->flags |= STA_STATE_FLAG_CONNECTED;
        }
    }

    pSta_info->state = info->sta_state;

    if (info->ssid_len) {
        pSta_info->ssid_len = info->ssid_len;
        memcpy(pSta_info->ssid, info->ssid, info->ssid_len);
        memcpy(pSta_info->bssid, info->bssid, IEEE80211_ADDR_LEN);
    }

    iRC = info->shim.shim_ops->get_beacon_interval(info, &beacon_interval);
    if (iRC) {
        beacon_interval = 0;
    }
    pSta_info->beacon_int = beacon_interval;

    iRC = info->shim.shim_ops->get_beacon_capabilities(info, &beacon_capabilities);
    if (iRC) {
        beacon_capabilities = 0;
    }
    pSta_info->beacon_capa = beacon_capabilities;

    iRC = info->shim.shim_ops->get_channel(info, &channel);
    if (iRC) {
        channel = 0;
    }
    pSta_info->freq = pSta_info->ap_freq = channel;

    if (copy_to_user((void *)args[1], pSta_info, sizeof(struct simple80211_sta_state_info))) {
        kfree(pSta_info);
        return -EFAULT;
    }

    kfree(pSta_info);
    return 0;
}

int simple80211_ioctl_get_open_mode(struct simple80211_info *info,
                                    unsigned long *args)
{
    unsigned int open_mode;

    open_mode = info->open_mode;
    if (copy_to_user((void *)args[1], &open_mode, sizeof(open_mode))) {
        return -EFAULT;
    }
    return 0;
}

int simple80211_ioctl_cache_open_mode(struct simple80211_info *info,
                                      unsigned long *args)
{
    unsigned int *pOpen_mode;

    pOpen_mode = (unsigned int *)args[1];
    info->open_mode = (*pOpen_mode == 1);
    return 0;
}

#ifdef SONOS_ARCH_ATTR_SUPPORTS_SUSPEND
int simple80211_ioctl_set_ip_addr(struct simple80211_info *info,
                                  unsigned long *args)
{
    info->ip_addr = (u_int32_t)args[1];
    return 0;
}
#endif

int simple80211_ioctl_scan_start(struct simple80211_info *info,
                                 unsigned long *args)
{
    int iRC;

    if (info->shim.shim_ops->scan_start) {
        iRC = info->shim.shim_ops->scan_start(info, args);
        if (0 == iRC) {
            printk("simple80211: scan started\n");
        }
        return iRC;
    }
    return -EINVAL;
}

int simple80211_ioctl_scan_abort(struct simple80211_info *info)
{
    if (info->shim.shim_ops->scan_abort) {
        return info->shim.shim_ops->scan_abort(info);
    }
    return -EINVAL;
}

int simple80211_ioctl_scan_get_results(struct simple80211_info *info,
                                       unsigned long *args)
{
    struct simple80211_scan_results *pScan_results;
    int rc;

    pScan_results = kmalloc(sizeof(struct simple80211_scan_results), GFP_KERNEL);
    if (!pScan_results) {
        return -ENOMEM;
    }
    spin_lock_bh(&info->info_lock);
    memcpy(pScan_results, &info->scan_info.scan_results, sizeof(*pScan_results));
    pScan_results->active = !!(info->scan_info.scan_flags & SCAN_FLAG_RUNNING);
    spin_unlock_bh(&info->info_lock);
    rc = copy_to_user((void *)args[1], pScan_results, sizeof(*pScan_results));
    if (rc) {
        rc = -EFAULT;
    }
    kfree(pScan_results);
    return rc;
}
int simple80211_ioctl_scan_clear(struct simple80211_info *info)
{
    spin_lock_bh(&info->info_lock);
    memset(&info->scan_info.scan_results, 0, sizeof(info->scan_info.scan_results));
    if (info->shim.shim_ops->scan_clear) {
        info->shim.shim_ops->scan_clear(info);
    }
    spin_unlock_bh(&info->info_lock);
    return 0;
}

#ifdef SONOS_ARCH_ATTR_SUPPORTS_SUSPEND
static inline int simple80211_ioctl_get_resume_reason(struct simple80211_info *info,
                                                      unsigned long *args)
{
    if (copy_to_user((void *)args[1], &info->resume_reason, sizeof(info->resume_reason))) {
        return -EFAULT;
    }
    return 0;
}

#ifdef SONOS_ARCH_ATTR_SUPPORTS_WOSSID
static inline int simple80211_ioctl_add_wossid(struct simple80211_info *info,
                                               unsigned long *args)
{
    struct simple80211_ssid *ssid = NULL;

    spin_lock_bh(&info->info_lock);
    if (info->wossid_list_len < WOSSID_MAX_SSID) {
        ssid = &info->wossid_list[info->wossid_list_len];
        memset(ssid, 0, sizeof(*ssid));
        ssid->ssid_len = (u_int8_t)args[2];
        ++info->wossid_list_len;
    } else {
        printk("WoSSID list full\n");
    }
    spin_unlock_bh(&info->info_lock);

    if (ssid == NULL) {
        return -EINVAL;
    }

    if (copy_from_user(ssid->ssid, (void *)args[1], sizeof(ssid->ssid))) {
        return -EFAULT;
    }

    return 0;
}

static inline int simple80211_ioctl_clr_wossid(struct simple80211_info *info,
                                               unsigned long *args)
{
    (void)args;
    spin_lock_bh(&info->info_lock);
    memset(&info->wossid_list[0], 0, sizeof(info->wossid_list));
    info->wossid_list_len = 0;
    spin_unlock_bh(&info->info_lock);
    return 0;
}

static inline void simple80211_start_wossid(struct simple80211_info *info)
{
    /* If WoSSID is configured, not connected to AP and not seeing APs, start WoSSID. */
    int start_wossid;

    spin_lock_bh(&info->info_lock);
    start_wossid = (info->sta_state != SIMPLE80211_STA_STATE_CONNECTED &&
                    info->wossid_list_len && info->shim.shim_ops->start_wossid &&
                    info->wossid_wake_loop_count < WOSSID_WAKE_LOOP_THRESHOLD &&
                    (info->base_dev->flags & IFF_UP) != 0);
    info->wossid_started = start_wossid;

    printk("%s: state: %d len: %d count: %d enable: %d\n",
           __func__, info->sta_state, info->wossid_list_len, info->wossid_wake_loop_count, start_wossid);

    if (!start_wossid) {
         /* With wireless enabled but no need to start WoSSID, then no need to keep counting. */
        info->wossid_wake_loop_count = 0;
    } else {
        do_gettimeofday(&info->wossid_suspend_time);
        if (info->wossid_wake_loop_count != 0) {
            struct timeval now;
            unsigned long interval;
            do_gettimeofday(&now);
            interval = now.tv_sec - info->wossid_resume_time.tv_sec;
            if (interval > WOSSID_WAKE_LOOP_SUSPEND_INTERVAL) {
                /* If it has been long enough since resume, we don't need to avoid the loop. */
                printk("%s: interval %lu. reset WoSSID counter\n", __func__, interval);
                info->wossid_wake_loop_count = 0;
            }
        }
    }

    spin_unlock_bh(&info->info_lock);

    if (start_wossid) {
        info->shim.shim_ops->start_wossid(info);
    }
}

static inline void simple80211_stop_wossid(struct simple80211_info *info, uint8_t resume_reason)
{
    spin_lock_bh(&info->info_lock);
    if (info->wossid_started && info->wossid_list_len &&
        info->shim.shim_ops->stop_wossid && resume_reason == SIMPLE80211_RESUME_NONE) {
        /* Stop the WoSSID if this is resume routine. */
        info->wossid_started = 0;
        info->shim.shim_ops->stop_wossid(info);
    } else if (resume_reason == SIMPLE80211_RESUME_WOSSID) {
        struct timeval now;
        unsigned long interval;

        do_gettimeofday(&now);
        interval = now.tv_sec - info->wossid_suspend_time.tv_sec;
        /* Update variables for WoSSID loop detecting if WoSSID is the resume reason. */
        if (interval < WOSSID_WAKE_LOOP_SCAN_INTERVAL) {
            info->wossid_wake_loop_count++;
            do_gettimeofday(&info->wossid_resume_time);
        } else {
            /* Waited long enough before finding AP, clear the loop counter. */
            info->wossid_wake_loop_count = 0;
        }
        printk("%s: stop WoSSID. interval: %lu count: %d\n",
               __func__, interval, info->wossid_wake_loop_count);
    }
    spin_unlock_bh(&info->info_lock);
}
#endif
#endif

int simple80211_ioctl(struct simple80211_info *info,
                      struct simple80211_cmd *cmd)
{
    int ret = -EINVAL;
    switch (cmd->data.ioctl.args[0]) {
        case SIMPLE80211_CTL_HAS_WIRELESS_CNX: {
            ret = simple80211_ioctl_has_wireless_cnx(info, cmd->data.ioctl.args);
            break;
        }
        case SIMPLE80211_CTL_SET_HHID: {
            ret = simple80211_ioctl_set_hhid(info, cmd->data.ioctl.args);
            break;
        }
        case SIMPLE80211_CTL_SET_CHANNEL: {
            ret = simple80211_ioctl_set_channel(info, cmd->data.ioctl.args);
            break;
        }
        case SIMPLE80211_CTL_GET_CHANNEL: {
            ret = simple80211_ioctl_get_channel(info, cmd->data.ioctl.args);
            break;
        }
        case SIMPLE80211_CTL_STA_SET_ENABLE: {
            ret = simple80211_ioctl_sta_set_enable(info, cmd->data.ioctl.args);
            break;
        }
        case SIMPLE80211_CTL_ADD_TXRX_FILTER: {
            ret = simple80211_ioctl_add_txrx_filter(info, cmd->data.ioctl.args);
            break;
        }
        case SIMPLE80211_CTL_STA_GET_STATE: {
            ret = simple80211_ioctl_sta_get_state(info, cmd->data.ioctl.args);
            break;
        }
        case SIMPLE80211_CTL_CACHE_OPEN_MODE: {
            ret = simple80211_ioctl_cache_open_mode(info, cmd->data.ioctl.args);
            break;
        }
        case SIMPLE80211_CTL_GET_OPEN_MODE: {
            ret = simple80211_ioctl_get_open_mode(info, cmd->data.ioctl.args);
            break;
        }
#ifdef SONOS_ARCH_ATTR_SUPPORTS_SUSPEND
        case SIMPLE80211_CTL_SET_IP_ADDR: {
            ret = simple80211_ioctl_set_ip_addr(info, cmd->data.ioctl.args);
            break;
        }
        case SIMPLE80211_CTL_GET_RESUME_REASON: {
            ret = simple80211_ioctl_get_resume_reason(info, cmd->data.ioctl.args);
            break;
        }
#ifdef SONOS_ARCH_ATTR_SUPPORTS_WOSSID
        case SIMPLE80211_CTL_WOSSID_ADD: {
            ret = simple80211_ioctl_add_wossid(info, cmd->data.ioctl.args);
            break;
        }
        case SIMPLE80211_CTL_WOSSID_CLR: {
            ret = simple80211_ioctl_clr_wossid(info, cmd->data.ioctl.args);
            break;
        }
#endif
#endif
        case SIMPLE80211_CTL_SCAN_START: {
            ret = simple80211_ioctl_scan_start(info, cmd->data.ioctl.args);
            break;
        }
        case SIMPLE80211_CTL_SCAN_ABORT: {
            ret = simple80211_ioctl_scan_abort(info);
            break;
        }
        case SIMPLE80211_CTL_SCAN_CLEAR: {
            ret = simple80211_ioctl_scan_clear(info);
            break;
        }
        case SIMPLE80211_CTL_SCAN_GET_RESULTS: {
            ret = simple80211_ioctl_scan_get_results(info, cmd->data.ioctl.args);
            break;
        }
        case SIMPLE80211_CTL_GET_LINK_STATS: {
            ret = simple80211_ioctl_get_link_stats(info, cmd->data.ioctl.args);
            break;
        }
        default:
            printk("%s: Unknown command type %lu\n", __func__, cmd->data.ioctl.args[0]);
            break;
    }
    return ret;
}

int simple80211_handle_add_key(struct simple80211_info *info,
                               struct simple80211_cmd *cmd)
{
    spin_lock_bh(&info->info_lock);
    if (info->sta_state == SIMPLE80211_STA_STATE_ASSOC_DONE) {
        if (cmd->data.key.cipher == WLAN_CIPHER_SUITE_WEP40 ||
            cmd->data.key.cipher == WLAN_CIPHER_SUITE_WEP104) {
            /* If it's WEP set both, but we may need to take a look whether
             * new chipsets need 2 WEP keys for UC and BC/MC.
             */
            info->sta_cipher_pairwise = info->sta_cipher_group = cmd->data.key.cipher;
            if (simple80211_in_credcheck(info)) {
                simple80211_nl_event(info, 0, NULL, 0, simple80211_renew_dhcp_event);
                printk("Requesting IP\n");
                spin_unlock_bh(&info->info_lock);
                return 0;
            }
        } else {
            if (cmd->data.key.is_pair) {
                info->sta_cipher_pairwise = cmd->data.key.cipher;
            } else {
                info->sta_cipher_group = cmd->data.key.cipher;
            }
        }
        /* Not WEP and keys are ready, directly call simple80211_handle_connected().  */
        if (info->sta_cipher_pairwise != WLAN_CIPHER_SUITE_INVALID &&
            info->sta_cipher_group != WLAN_CIPHER_SUITE_INVALID) {
            simple80211_handle_connected(info);
        }
    }
    spin_unlock_bh(&info->info_lock);
    return 0;
}

#define residual_len(skb, curr) \
    ((skb)->len - ((unsigned char *)(curr) - (unsigned char *)(skb)->data))

int simple80211_txrx_filter(struct simple80211_info *info,
                            struct sk_buff *skb, unsigned char is_tx)
{
    struct ethhdr *eh;
    struct udphdr *udph = NULL;
    struct tcphdr *tcph = NULL;
    struct iphdr *ipv4h = NULL;
    struct ipv6hdr *ipv6h = NULL;
    struct simple80211_txrx_filter_node *filter_node;
    unsigned short ether_type;
    unsigned char check_ethertype = 0;
    unsigned char is_ipv4;
    unsigned char is_ipv6;
    unsigned char *next_hdr;
    unsigned char allowed = 0;
    unsigned short sport = 0, dport = 0;
    unsigned char ipproto = 0;

    if (skb->len < sizeof(struct ethhdr)) {
        return -EPROTO;
    }

    eh = (struct ethhdr *)skb_mac_header(skb);
    ether_type = ntohs(eh->h_proto);
    is_ipv4 = (ether_type == ETH_P_IP);
    is_ipv6 = (ether_type == ETH_P_IPV6);
    next_hdr = (unsigned char *)eh + sizeof(struct ethhdr);

    if (!is_ipv4 && !is_ipv6) {
        /* If it's not IPv4 or IPv6, we have to check the ether type. */
        check_ethertype = 1;
    } else {
        /* Get networking layer info */
        if (is_ipv4) {
            if (residual_len(skb, next_hdr) < sizeof(struct iphdr)) {
                return -EPROTO;
            }
            ipv4h = (struct iphdr *)next_hdr;
            if (ipv4h->version != 4) {
                return -EPROTO;
            }
            if (residual_len(skb, next_hdr) < (ipv4h->ihl * 4)) {
                return -EPROTO;
            }
            if (ipv4h->frag_off != 0 && is_tx) {
                /* Allow non-initial IP fragments for TX. */
                return 0;
            }
            next_hdr += (ipv4h->ihl * 4);
            ipproto = ipv4h->protocol;
        } else {
            /* Is IPv6 */
            if (residual_len(skb, next_hdr) < sizeof(struct ipv6hdr)) {
                return -EPROTO;
            }
            ipv6h = (struct ipv6hdr *)next_hdr;
            if (ipv6h->version != 6) {
                return -EPROTO;
            }
            next_hdr += sizeof(struct ipv6hdr);
            ipproto = ipv6h->nexthdr;
            if (ipproto == NEXTHDR_HOP) {
                struct ipv6_opt_hdr *hp;
                int hdr_len;
                int offset = sizeof(struct ipv6hdr);

                if (residual_len(skb, next_hdr) < sizeof(struct ipv6_opt_hdr)) {
                    return -EPROTO;
                }
                hp = (struct ipv6_opt_hdr *)next_hdr;
                ipproto = hp->nexthdr;
                hdr_len = ipv6_optlen(hp);

                if (residual_len(skb, next_hdr) < hdr_len) {
                    return -EPROTO;
                }
                offset += hdr_len;
                next_hdr += hdr_len;
            }
        }

        /* Get transport layer info */
        if (ipproto == IPPROTO_UDP) {
            if (residual_len(skb, next_hdr) < sizeof(struct udphdr)) {
                return -EPROTO;
            }
            udph = (struct udphdr *)next_hdr;
            sport = ntohs(udph->source);
            dport = ntohs(udph->dest);
            next_hdr += sizeof(struct udphdr);
        } else if (ipproto == IPPROTO_TCP) {
            unsigned int hdr_len;
            if (residual_len(skb, next_hdr) < sizeof(struct tcphdr)) {
                return -EPROTO;
            }
            tcph = (struct tcphdr *)next_hdr;
            hdr_len = tcph->doff * 4;
            if (residual_len(skb, next_hdr) < hdr_len) {
                return -EPROTO;
            }
            sport = ntohs(tcph->source);
            dport = ntohs(tcph->dest);
            next_hdr += hdr_len;
        }
    }

    /* Now we got ether type, IP proto, ports and address of the payload part.
     * Loop through the filters now.
     */
    spin_lock_bh(&info->filter_lock);
    list_for_each_entry(filter_node, &info->txrx_filter_list, list) {
        unsigned char mirror;
        if (allowed && (filter_node->filter.flags & FILTER_ALLOW) &&
            !(filter_node->filter.flags & FILTER_MANGLE)) {
            /* Already allowed and nothing to mangle. */
            continue;
        }
        /* Only set mirror to be true if this filter really cares about port. */
        mirror = ((filter_node->filter.flags & (FILTER_DPORT | FILTER_SPORT)) &&
                  (filter_node->filter.flags & FILTER_MIRROR));
        if ((is_tx && (filter_node->filter.flags & FILTER_TX)) ||
            (!is_tx && (filter_node->filter.flags & FILTER_RX))) {
            /* Direction of the current packet matches this filter. If this filter is not
             * bi-directional, we should disable mirror.
             */
            mirror = (mirror && ((filter_node->filter.flags & FILTER_TXRX) == FILTER_TXRX));
        } else if (!mirror) {
            /* Not going to do mirror on this filter and mis-matched direction. Skip filter. */
            continue;
        }
        if (!check_ethertype != !(filter_node->filter.flags & FILTER_ETHERTYPE)) {
            continue;
        }
        /* Matched direction, ether type and not repeating checking allowed packet against
         * the policy that doesn't drop it.
         */
        if (check_ethertype) {
            if (ether_type != filter_node->filter.ethertype) {
                continue;
            }
        } else {
            /* This is a IP packet. Get the ports and proto. */
            if ((is_ipv4 && (filter_node->filter.flags & FILTER_IPV4)) ||
                (is_ipv6 && (filter_node->filter.flags & FILTER_IPV6))) {
                if ((filter_node->filter.flags & FILTER_IPPROTO) &&
                    filter_node->filter.ipproto != ipproto) {
                    continue;
                }
                if ((ipproto == IPPROTO_TCP || ipproto == IPPROTO_UDP)) {
                    /* Use the source port as destination port and use destination port as source port
                     * if mirror flag is set.
                     */
                    if ((filter_node->filter.flags & FILTER_SPORT) &&
                        ((filter_node->filter.sport != sport) &&
                         (!mirror || filter_node->filter.sport != dport))) {
                        continue;
                    }
                    if ((filter_node->filter.flags & FILTER_DPORT) &&
                        ((filter_node->filter.dport != dport) &&
                         (!mirror || filter_node->filter.dport != sport))){
                        continue;
                    }
                }
            } else {
                /* The filter didn't match the IP header version. */
                continue;
            }
        }
        /* Reaching here means this packet matched this filter for everything except payload part.
         * Check the payload now.
         */
        if (filter_node->filter.flags & FILTER_MATCH) {
            if (residual_len(skb, next_hdr) <= filter_node->filter.match.offset) {
                /* Unlike cases that don't have enough length where we have to drop the packet.
                 * Hitting here just means the packet is not long enough to match the pattern.
                 */
                continue;
            }
            if ((next_hdr[filter_node->filter.match.offset] & filter_node->filter.match.mask) != filter_node->filter.match.value) {
                continue;
            }
        }
        /* Matched everything. */
        if (filter_node->filter.flags & FILTER_DROP) {
            allowed = 0;
            break;
        }
        if ((filter_node->filter.flags & FILTER_MANGLE) && (ipv4h != NULL) &&
            ((udph != NULL) || (tcph != NULL))) {
            /* Only mangle IPv4 UDP payload for now. */
            unsigned char new_value;
            unsigned short tcpudp_len;
            unsigned char *csum_start;
            u_int16_t *csum;

            if (residual_len(skb, next_hdr) <= filter_node->filter.mangle.offset) {
                /* We can't mangle this packet as expected. Just skip this filter without
                 * marking it as allowed and see whether we can hit any other filter.
                 */
                continue;
            }
            new_value = next_hdr[filter_node->filter.mangle.offset];
            new_value &= ~filter_node->filter.mangle.mask;
            new_value |= filter_node->filter.mangle.value;
            next_hdr[filter_node->filter.mangle.offset] = new_value;
            tcpudp_len = residual_len(skb, next_hdr);
            if (udph) {
                csum_start = (unsigned char *)udph;
                udph->check = 0;
                csum = &udph->check;
                tcpudp_len += sizeof(struct udphdr);
            } else {
                /* tcph should not be NULL then. */
                csum_start = (unsigned char *)tcph;
                tcph->check = 0;
                csum = &tcph->check;
                tcpudp_len += (tcph->doff * 4);
            }
            *csum = csum_tcpudp_magic(ipv4h->saddr, ipv4h->daddr,
                                      tcpudp_len, ipproto,
                                      csum_partial(csum_start, tcpudp_len, 0));
        }
        allowed = 1;
    }
    spin_unlock_bh(&info->filter_lock);
    return (allowed ? 0 : -EPROTO);
}

int simple80211_handle_scan_event(struct simple80211_info *info,
                                  struct simple80211_cmd *cmd)
{
    if (SIMPLE80211_SCAN_EVENT_PRESCAN == cmd->data.scan_event.event_type) {
        /* For now we are not doing anything with this event.  If we need to use it later,
         * we can use spin_lock for synchronization as HW interrupts can happen before or during this call.
         */
        return 0;
    }
    spin_lock_bh(&info->info_lock);
    if (SIMPLE80211_SCAN_EVENT_SCANNING == cmd->data.scan_event.event_type) {
        info->scan_info.scan_flags |= SCAN_FLAG_RUNNING;
        info->scan_info.scan_results.num_aps = 0;
    } else if ((SIMPLE80211_SCAN_EVENT_SCAN_DONE == cmd->data.scan_event.event_type) ||
            (SIMPLE80211_SCAN_EVENT_SCAN_ABORTED == cmd->data.scan_event.event_type)) {
        ++info->sta_scan_count;
        if (info->shim.shim_ops->scan_save_results) {
            info->shim.shim_ops->scan_save_results(info);
        }
        if ((SIMPLE80211_SCAN_EVENT_SCAN_DONE == cmd->data.scan_event.event_type) &&
                (info->scan_info.scan_flags & SCAN_FLAG_UMAC_SCAN)) {
            printk("simple80211: scan is done\n");
        }
        info->scan_info.scan_flags = 0;
    }
    spin_unlock_bh(&info->info_lock);
    return 0;
}

int simple80211_handle_cmd(struct simple80211_info *info,
                           struct simple80211_cmd *cmd)
{
    int ret = -1;

    if (cmd == NULL || info == NULL) {
        printk("%s: error input info: %p cmd: %p\n", __func__, info, cmd);
        dump_stack();
        return ret;
    }

    switch (cmd->cmd_type) {
        case SIMPLE80211_INIT_INFO: {
            ret = simple80211_init_info(info, cmd);
            break;
        }
        case SIMPLE80211_BR_UPLINK: {
            ret = simple80211_uplink(info, cmd);
            break;
        }
        case SIMPLE80211_IOCTL: {
            ret = simple80211_ioctl(info, cmd);
            break;
        }
        case SIMPLE80211_ADD_KEY: {
            ret = simple80211_handle_add_key(info, cmd);
            break;
        }
        case SIMPLE80211_TXRX_FILTER: {
            ret = simple80211_txrx_filter(info, cmd->data.txrx_filter.skb,
                                          cmd->data.txrx_filter.is_tx);
            break;
        }
        case SIMPLE80211_SCAN_EVENT: {
            ret = simple80211_handle_scan_event(info, cmd);
            break;
        }
        case SIMPLE80211_STA_GET_STATE: {
            ret = simple80211_sta_get_state(info, cmd);
            break;
        }
#ifdef SONOS_ARCH_ATTR_SUPPORTS_SUSPEND
        case SIMPLE80211_SUSPEND: {
            info->resume_reason = SIMPLE80211_RESUME_NONE;
            info->shim.shim_ops->pwr_suspend(info);
#ifdef SONOS_ARCH_ATTR_SUPPORTS_WOSSID
            simple80211_start_wossid(info);
#endif
            ret = 0;
            break;
        }
        case SIMPLE80211_RESUME: {
            if (info->resume_reason == SIMPLE80211_RESUME_NONE &&
                cmd->data.power_state.resume_reason != SIMPLE80211_RESUME_NONE) {
                /* Only update resume reason if it's valid. */
                info->resume_reason = cmd->data.power_state.resume_reason;
            } else if (cmd->data.power_state.resume_reason == SIMPLE80211_RESUME_NONE) {
                /* Make sure pwr_resume() is only called for once. */
                info->shim.shim_ops->pwr_resume(info);
            }
#ifdef SONOS_ARCH_ATTR_SUPPORTS_WOSSID
            simple80211_stop_wossid(info, cmd->data.power_state.resume_reason);
#endif
            ret = 0;
            break;
        }
#endif
        default:
            printk("%s: Unknown command type %u\n", __func__, cmd->cmd_type);
    }
    return ret;
}

static int __init init_simple80211(void)
{
    printk("%s: Installed simple80211\n", __func__);
    return 0;
}
module_init(init_simple80211);

static void __exit exit_simple80211(void)
{
    printk("%s: Removed simple80211\n", __func__);
}
module_exit(exit_simple80211);

MODULE_AUTHOR("Sonos, Inc.");
MODULE_DESCRIPTION("Helper module for WiFi");
MODULE_LICENSE("GPL");
